{-# LANGUAGE TypeFamilies #-}
module Section34NativeLet where

-- The TypeFamilies extension is required to instantiate the 'MuRef'
-- class for data-reify.

import Control.Applicative (pure, (<$>), (<*>))
import Data.Maybe (fromJust)
import Data.Reify (MuRef(..), reifyGraph, Graph(..))
import Section3Recursion

-- Like ExprD from Section2Sharing, but using LitD instead of OneD:
data ExprD = LitD Int | AddD ExprD ExprD
  deriving (Show)

oneD :: ExprD
oneD = LitD 1

(.+.!) :: ExprD -> ExprD -> ExprD
(.+.!) = AddD

treeD_I :: Int -> ExprD
treeD_I 0 = oneD
treeD_I n = let shared = treeD_I (n - 1) in shared .+.! shared

data ExprF r = LitF Int | AddF r r
  deriving (Show)

instance MuRef ExprD where
  type DeRef ExprD = ExprF

  mapDeRef f (LitD n)     = pure (LitF n)
  mapDeRef f (AddD e1 e2) = AddF <$> f e1 <*> f e2

test_reify_Graph = reifyGraph (treeD_I 1)

build :: [(Int, ExprF Int)] -> Int -> ClosedExpr
build env root =
  letrec_  (\ vs ->  let  go (LitF x)      =  Lit x
                          go (AddF v1 v2)  =
                            Add (var vs v1) (var vs v2)
                     in map (go . snd) env)
           (\ vs -> var vs root)
  where
    var vs n =
      fromJust (lookup n (zipWith (\ (i, _) x -> (i, x)) env vs))

test = do (Graph env r) <- reifyGraph (treeD_I 3)
          print (text (build env r))
