{-# LANGUAGE RankNTypes, KindSignatures, GADTs #-}
module Section4TypedLists where

-- More type system extensions are required for this part of the
-- paper. In particular, we need GADTs.

data TList :: (* -> *) -> * -> * where
  TNil  :: TList f ()
  (:::) :: f t -> TList f ts -> TList f (t,ts)

infixr 7 :::

-- Heterogeneous and homogeneous lists

newtype I a = I { unI :: a }
  deriving (Show)

hlist :: TList I (Int, (Int -> Int, (Bool, ())))
hlist = I 3 ::: I (\x -> x) ::: I False ::: TNil

newtype K b a = K { unK :: b }
  deriving (Show)

list :: TList (K Int) (t, (t1, (t2, ())))
list = K 1 ::: K 2 ::: K 3 ::: TNil

-- Basic operations

thead :: TList f (t,ts) -> f t
thead (x ::: xs) = x

ttail :: TList f (t,ts) -> TList f ts
ttail (x ::: xs) = xs

tlength :: TList v t -> Int
tlength TNil        = 0
tlength (x ::: xs)  = 1 + tlength xs

-- Mapping and zipping

tmap :: (forall t. f t -> g t) -> TList f t -> TList g t
tmap f TNil       = TNil
tmap f (x ::: xs) = f x ::: tmap f xs

tmapK :: (a -> b) -> TList (K a) ts -> TList (K b) ts
tmapK f = tmap (K . f . unK)

tzipWith :: (forall t . f t -> g t -> h t) ->
            TList f ts -> TList g ts -> TList h ts
tzipWith f TNil       TNil       = TNil
tzipWith f (x ::: xs) (y ::: ys) = f x y ::: tzipWith f xs ys

-- Producers of typed lists

data RList :: * -> * where
  RNil  :: RList ()
  RCons :: RList ts -> RList (t, ts)

titerate' :: RList ts -> (a -> a) -> a -> TList (K a) ts
titerate' RNil f n       = TNil
titerate' (RCons xs) f n = K n ::: titerate' xs f (f n)

-- Using type classes for producers

titerate :: CList ts => (a -> a) -> a -> TList (K a) ts
titerate = titerate' cList

class CList t where
  cList :: RList t

instance CList () where
  cList = RNil

instance CList ts => CList (t, ts) where
  cList = RCons cList

tenumFrom :: CList ts => Int -> TList (K Int) ts
tenumFrom n = titerate (+1) n

test :: TList (K Int) (t1, (t2, ()))
test = tenumFrom 0
